#ifndef LIBDB_H
#define LIBDB_H

/* Includes */

#include <sys/types.h>
#include <stdint.h>

/* Macros */

/**
 * Signals that the DB needs to be recreated. For use with
 * #libsw_db_api#open function.
 */
#define LIBSW_DB_REBUILD 1

/**
 * The successful return value.
 */
#define LIBSW_DB_OK 0

/**
 * The unsuccessful return value.
 */
#define LIBSW_DB_ERROR 1

/**
 * The special return value signalling the loader function to stop the
 * data loading (see #libsw_db#load_data).
 */
#define LIBSW_DB_LOADER_STOP 255

/* Types */

/**
 * DB initialization callback type (a data record loader). The
 * #LIBSW_DB_LOADER_STOP return value is used to signal the calling DB
 * loader that there are no more records to load.
 *
 * @param record the record pointer to load the data in
 * @param index the record index
 * @param ldctx the loader context
 *
 * @return 0 on success or \c LIBSW_DB_LOADER_STOP on success, any
 * other value on fail
 */
typedef int (*libsw_db_loader_t)
            ( void *record, uint16_t index, void *ldctx );


/**
 * Database API descriptor / handle.
 */
typedef struct libsw_db {

  /**
   * Free to use context pointer.
   */
  void *ctx;

  /**
   * Calls the given loader function for each record. The loading
   * continues until \c db->#get_count database records are loaded or
   * the loader function returns non-zero. Returning the
   * #LIBSW_DB_LOADER_STOP value stops the loading but results in
   * successfull status for the overall operation.
   *
   * @param db DB handle
   * @param loader the loader function
   * @param ctx the loader context (may be NULL)
   *
   * @return 0 on success, non-zero on fail
   */
  int (*load_data) ( struct libsw_db *db,
                     libsw_db_loader_t loader, void *ctx );

  /**
   * Returns the record pointer for the given index or NULL,
   * if the specified index is beyond the DB length.
   *
   * @param db DB handle
   * @param index record index
   *
   * @return pointer to the record or NULL
   */
  void *(*get_record) ( struct libsw_db *db, uint16_t index );

  /**
   * Returns the index of the given record.
   *
   * @param db DB handle
   * @param record record pointer
   *
   * @return record index
   */
  uint16_t (*get_index) ( struct libsw_db *db, const void *record );

  /**
   * Returns the overall record count.
   *
   * @param db DB handle
   *
   * @return number of records in the DB
   */
  uint16_t (*get_count) ( struct libsw_db *db );

  /**
   * Puts the lock on the DB as a whole. The further calls to that
   * function are blocked until #unlock_db is called.
   *
   * @param db DB handle
   *
   * @return 0 on success, non-zero on fail
   */
  int (*lock_db) ( struct libsw_db *db );

  /**
   * Removes the lock from the DB as a whole. All modifications made
   * to the DB records are synchronized with the underlying storage.
   *
   * @param db DB handle
   *
   * @return 0 on success, non-zero on fail
   */
  int (*unlock_db) ( struct libsw_db *db );

   /**
   * Locks the given single DB record. The further calls to that
   * function are blocked until #unlock_record is called.
   *
   * @param db DB handle
   * @param record record pointer
   *
   * @return 0 on success, non-zero on fail
   */
  int (*lock_record) ( struct libsw_db *db, const void *record );

  /**
   * Unlocks the given single DB record. All modifications made
   * to the record are synchronized with the underlying storage.
   *
   * @param db DB handle
   * @param record record pointer
   *
   * @return 0 on success, non-zero on fail
   */
  int (*unlock_record) ( struct libsw_db *db, const void *record );

  /**
   * Closes this DB handle, freeing up the allocated resources. The DB
   * handle becomes unusable.
   * 
   * @param db DB handle
   *
   * @return 0 on success, non-zero on fail
   */
  int (*close) ( struct libsw_db *db );
} libsw_db_t;


/**
 * Database manager (DBA) API descriptor.
 */
typedef struct libsw_db_api {

  /**
   * Free to use context pointer.
   */
  void *ctx;

  /**
   * Returns the initialized DB handle for the database with given
   * parameters. If the described database doesn't exist it is created
   * and initialized with the help of the given \c loader function.
   * The DB re-creation can be forced by specifying the
   * #LIBSW_DB_REBUILD flag in \c flags.
   *
   * @param api DBA handle
   * @param name DB name
   * @param rcount number of records (DB szie)
   * @param rsize record length, bytes
   * @param flags various flags
   * @param loader the record loader function
   * @param ldctx the loader function context (any data or NULL)
   *
   * @return initialized DB access handle or NULL on error
   */
  libsw_db_t *(*open) ( struct libsw_db_api *api, const char *name,
                        uint16_t rcount, size_t rsize,
                        uint16_t flags,
                        libsw_db_loader_t loader,
                        void *ldctx );

  /**
   * Drops (deletes) the database with the given name. All data is
   * LOST!
   *
   * @param api DBA handle
   * @param name DB name
   *
   * @return initialized DB access handle or NULL on error
   */
  int (*drop) ( struct libsw_db_api *api, char *name );

  /**
   *
   * Disconnects from the DBA, closes this DBA handle, freeing up the
   * allocated resources. The DBA handle becomes unusable.
   * 
   * @param db DBA handle
   *
   * @return 0 on success, non-zero on fail
   */
  int (*close) ( struct libsw_db_api *api );

} libsw_db_api_t;


/* Functions (documented in libdb.c) */

libsw_db_api_t *libsw_get_db_api( const char *driver,
                                  const char *config );


#endif
