#include "libdb.h"
#include "libmemdb.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>

/* #define LIBDB_DEBUG */

#ifdef LIBDB_DEBUG
#define DBGL(level, format, ...) \
    printf("DBG: %s: "format"\n", __FUNCTION__, ##__VA_ARGS__)
#else
#define DBGL(level, format, ...)
#endif
#define DBG(format, ...) DBGL(0, format, ##__VA_ARGS__)

#define DBFNPAT "%s/%s.memdb"

/**
 * Helper funtion that makes all of the components of the given file
 * path except the last one.
 *
 * @param api DBA handle
 * @param filepath file path
 *
 * @return 0 on success, non-zero on fail
 */
int mkdirs( libsw_memdb_api_t *api, const char *filepath )
{
    char tmp[ 1024 ];
    size_t len;
    int i;

    struct stat st;

    snprintf( tmp, sizeof( tmp ), "%s", filepath );
    len = strlen( tmp );

    for( i = 1; (size_t)i < len; i++ ) {
      if( tmp[ i ] == '/' ) {
        tmp[ i ] = 0;
        if ( stat( tmp, &st ) != 0 ) {
          if ( mkdir( tmp, api->mode | S_IXUSR | S_IXGRP ) != 0 ) {
            return LIBSW_DB_ERROR;
          }
          if ( chown( tmp, api->uid, api->gid ) != 0 ) {
            return LIBSW_DB_ERROR;
          }
        }
        tmp[ i ] = '/';
      }
    }

    return LIBSW_DB_OK;
}

/**
 * Maps the given DB file into memory, optionally performing the
 * initialization procedure. The DB file is created if it doesn't
 * exist.
 *
 * If the DB file doesn't exist or the #LIBSW_DB_REBUILD flag is
 * given, then the file is initialized with records produced by the
 * given loader function (if any) or filled with zeros. For more
 * information see the description of #libsw_load_obj_db.
 *
 * @param api DBA handle
 * @param db Mem-DB descriptor (filename, size, ...)
 * @param flags DB file open flags
 * @param loader the loader function
 * @param ldctx the loader context (may be NULL)
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_open_obj_db( libsw_memdb_api_t *api, libsw_obj_db_t *db,
                       uint16_t flags,
                       libsw_db_loader_t loader, void *ldctx )
{
  size_t dbsize = db->count * db->isize;
  int oflags = O_RDWR | O_CREAT;
  struct stat st;

  DBG( "%p { n : \"%s\", size : %u, flags : %u }",
       db, db->filename, dbsize, flags );

  if ( stat( db->filename, &st ) != 0 ) {
    flags |= LIBSW_DB_REBUILD;
    if ( mkdirs( api, db->filename ) != LIBSW_DB_OK )
      goto error;
  }

  if ( flags & LIBSW_DB_REBUILD )
    oflags |=  O_TRUNC;

  db->fd = open( db->filename, oflags );
  if ( db->fd == - 1 )
    goto error;

  if ( flags & LIBSW_DB_REBUILD ) {
    /* Set access mode for new/rebuilt files only */
    if ( fchown( db->fd, api->uid, api->gid ) != 0 )
        goto error;
    if ( fchmod( db->fd, api->mode ) != 0 )
        goto error;
  }

  if ( flags & LIBSW_DB_REBUILD ) {
    if ( ftruncate( db->fd, 0 ) != 0 )
      goto error;
    if ( ftruncate( db->fd, dbsize ) != 0 )
      goto error;
  } 

  db->head = mmap( NULL, dbsize, PROT_READ | PROT_WRITE ,
                   MAP_SHARED, db->fd, 0 );

  if ( flags & LIBSW_DB_REBUILD ) {

    memset( db->head, 0, dbsize );
    if ( write( db->fd , db->head, dbsize ) != (ssize_t)dbsize )
      goto error;
  }

  if ( flags & LIBSW_DB_REBUILD && loader != NULL ) {
    if ( libsw_load_obj_db( db, loader, ldctx ) != LIBSW_DB_OK )
      goto error;
  }

  if ( db->head == MAP_FAILED )
    goto error;


  return LIBSW_DB_OK;


error:

  if ( ( db->head != MAP_FAILED ) && db->head )
    munmap( db->head, dbsize );

  if ( db->fd != -1 ) {
    close( db->fd );
    db->fd = -1;
  }


  return LIBSW_DB_ERROR;

}


/**
 * Calls the given loader function for each record. The loading
 * continues until \c db->count database records are loaded or the
 * loader function returns non-zero. Returning the
 * #LIBSW_DB_LOADER_STOP value stops the loading but results in
 * successfull status for the overall operation.
 *
 * @param db Mem-DB descriptor
 * @param loader the loader function
 * @param ldctx the loader context (may be NULL)
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_load_obj_db( libsw_obj_db_t *db, libsw_db_loader_t loader,
                       void *ldctx )
{
  int ret = LIBSW_DB_OK;
  int i;

  DBG( "%p { n : \"%s\", loader : %p, ldctx : %p }",
       db, db->filename, loader, ldctx );

  if ( ! loader )
     return LIBSW_DB_OK;
  
  libsw_lock_obj_db( db );

  for ( i = 0; i < db->count; i++ ) {
    ret = loader( libsw_get_obj_db_item( db, i ), i, ldctx );
    if ( ret != LIBSW_DB_OK )
      break;
  }

  libsw_unlock_obj_db( db );

  if ( ret == LIBSW_DB_LOADER_STOP ) {
    ret = LIBSW_DB_OK;
  }

  return ret;
}


/**
 * Returns the object record pointer for the given offset or NULL,
 * if the specified offset is beyond DB length.
 *
 * @param db Mem-DB handle
 * @param idx object index 
 *
 * @return object record pointer
 */
void *libsw_get_obj_db_item( libsw_obj_db_t *db, uint16_t idx )
{

  DBG( "( %p { n : \"%s\", head : %p, isize : %u, count : %u }, "
       "idx = %u )",
       db, db->filename, db->head, db->isize, db->count, idx );

  if ( idx >= db->count ) {
    DBG( "  -> DB index out of bounds: %u > %u", idx, db->count );
    return NULL;
  } else {
    uint8_t* raw_head = (uint8_t*) db->head;
    uint32_t raw_offs = db->isize * idx;
    uint8_t* res = raw_head + raw_offs;

    DBG( "  -> %p + %u = %p", raw_head, raw_offs, res );

    return res;
  }
}


/**
 * Returns the offset of the given object record.
 *
 * @param db Mem-DB handle
 * @param item object record pointer
 *
 * @return object record offset
 */
uint16_t libsw_get_obj_db_index( libsw_obj_db_t *db,
                                 const void *item )
{
  DBG( "( %p { n : \"%s\", head : %p, isize : %u, count : %u }, "
       "item = %p )",
       db, db->filename, db->head, db->isize, db->count, item );

  uint8_t* raw_head = (uint8_t*) db->head;
  const uint8_t* raw_item = (const uint8_t*) item;
  uint16_t res = (raw_item - raw_head) / db->isize;

  DBG( "  -> (%p - %p) / %u = %u", raw_item, raw_head, db->isize, res );

  return res;
}


/**
 * Puts the lock on the DB. All further calls are blocked until
 * #libsw_unlock_obj_db is called.
 *
 * @param db Mem-DB handle
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_lock_obj_db( libsw_obj_db_t *db )
{
  if ( lseek( db->fd, 0, SEEK_SET ) == 0 ) {
    if ( lockf( db->fd, F_LOCK, 0 ) == 0 ) {
      return LIBSW_DB_OK;
    }
  }

  return LIBSW_DB_ERROR;
}

/**
 * Removes the lock from the DB. The whole Mem-DB contents are
 * synchronized with the mapped file.
 *
 * @param db Mem-DB handle
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_unlock_obj_db( libsw_obj_db_t *db )
{
  if ( lseek( db->fd, 0, SEEK_SET ) == 0 ) {
    if ( db->head )
      msync( db->head, db->isize * db->count, MS_INVALIDATE );
    if ( lockf( db->fd, F_ULOCK, 0 ) == 0 ) {
      return LIBSW_DB_OK;
    }
  }

  return LIBSW_DB_ERROR;
}

/**
 * Puts the lock on the DB record. All further calls are blocked until
 * #libsw_unlock_obj_db_item is called.
 *
 * @param db Mem-DB handle
 * @param item object record pointer
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_lock_obj_db_item( libsw_obj_db_t *db, const void *item )
{
  uint16_t inum = libsw_get_obj_db_index( db, item );
  if ( lseek( db->fd, inum * db->isize, SEEK_SET ) >= 0 ) {
    if ( lockf( db->fd, F_LOCK, db->isize ) == 0 ) {
      return LIBSW_DB_OK;
    }
  }

  return LIBSW_DB_ERROR;
}

/**
 * Removes the lock from the DB record. The corresponding portion of
 * the Mem-DB contents are synchronized with the mapped file part.
 *
 * @param db Mem-DB handle
 * @param item object record pointer
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_unlock_obj_db_item( libsw_obj_db_t *db, const void *item )
{    
  uint16_t inum =  libsw_get_obj_db_index( db, item );

  if ( db->head )
    msync( (void *)item, db->isize, MS_INVALIDATE );
  if ( lseek( db->fd, inum * db->isize, SEEK_SET ) >= 0 ) {
    if ( lockf(db->fd, F_ULOCK, db->isize) == 0 ) {
      return LIBSW_DB_OK;
    }
  }

  return LIBSW_DB_ERROR;
}


/**
 * Unmaps the memory mapped DB file and frees the allocated memory.
 *
 * @param db Mem-DB handle
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_close_obj_db( libsw_obj_db_t *db )
{
  size_t dbsize = db->isize * db->count;

  DBG( "%p { n : \"%s\", size : %u }", db, db->filename, dbsize );

  if ( db->fd != -1 ) {
    if ( db->head ) {
      msync( db->head, dbsize, MS_INVALIDATE );
      munmap( db->head, dbsize );
    }

    close( db->fd );
    db->fd = -1;
  }


  return LIBSW_DB_OK; //TODO: check for errors
}


/* DB access API functions */

/**
 * DB API wrapper of #libsw_load_obj_db.\ Implements DB API
 * #libsw_db#load_data function.
 */
int libsw_db_load_data( struct libsw_db *db,
                        libsw_db_loader_t loader, void *ldctx )
{
  return libsw_load_obj_db( (libsw_obj_db_t *) db->ctx,
                            loader, ldctx );
}

/**
 * DB API wrapper of #libsw_get_obj_db_item.\ Implements DB API
 * #libsw_db#get_record function.
 */
void *libsw_db_get_record( struct libsw_db *db, uint16_t index )
{
  return libsw_get_obj_db_item( (libsw_obj_db_t *) db->ctx, index );
}

/**
 * DB API wrapper of #libsw_get_obj_db_index.\ Implements DB API
 * #libsw_db#get_index function.
 */
uint16_t libsw_db_get_index( struct libsw_db *db,
                                 const void *record )
{
  return libsw_get_obj_db_index( (libsw_obj_db_t *) db->ctx, record );
}

/**
 * Implements DB API #libsw_db#get_count function.
 */
uint16_t libsw_db_get_count( struct libsw_db *db )
{
  libsw_obj_db_t *dbctx = (libsw_obj_db_t *) db->ctx;
  return dbctx->count;
}

/**
 * DB API wrapper of #libsw_lock_obj_db.\ Implements DB API
 * #libsw_db#lock_db function.
 */
int libsw_db_lock_db( struct libsw_db *db )
{
  return libsw_lock_obj_db( (libsw_obj_db_t *) db->ctx );
}

/**
 * DB API wrapper of #libsw_unlock_obj_db.\ Implements DB API
 * #libsw_db#unlock_db function.
 */
int libsw_db_unlock_db( struct libsw_db *db )
{
  return libsw_unlock_obj_db( (libsw_obj_db_t *) db->ctx );
}

/**
 * DB API wrapper of #libsw_lock_obj_db_item.\ Implements DB API
 * #libsw_db#lock_record function.
 */
int libsw_db_lock_record( struct libsw_db *db,
                              const void *record )
{
  return libsw_lock_obj_db_item( (libsw_obj_db_t *) db->ctx, record );
}

/**
 * DB API wrapper of #libsw_unlock_obj_db_item.\ Implements DB API
 * #libsw_db#unlock_record function.
 */
int libsw_db_unlock_record( struct libsw_db *db, const void *record )
{
  return libsw_unlock_obj_db_item( (libsw_obj_db_t *) db->ctx,
                                   record );
}

/**
 * DB API wrapper of #libsw_close_obj_db.\ Implements DB API
 * #libsw_db#close function. Additionally, frees the given DB handle,
 * making it unusable.
 */
int libsw_db_close( struct libsw_db *db )
{
  int ret = LIBSW_DB_OK;

  if ( db ) {
    if ( db->ctx ) {
      ret = libsw_close_obj_db( (libsw_obj_db_t *) db->ctx );
      free(db->ctx);
      db->ctx = NULL;
    }

    if ( ret == LIBSW_DB_OK ) {
      free( db );
    }
  }

  return ret;
}


/* DB manager API functions */


/**
 * Wraps call to the #libsw_open_obj_db.\ Implements DBA API
 * #libsw_db_api#open function.
 *
 * @return initialized DB access handle or NULL on error
 */
libsw_db_t *libsw_db_api_open( struct libsw_db_api *api,
                               const char *name,
                               uint16_t rcount,
                               size_t rsize,
                               uint16_t flags,
                               libsw_db_loader_t loader,
                               void *ldctx )
{
  libsw_db_t *db = NULL;
  libsw_obj_db_t *odb = NULL;
  libsw_memdb_api_t *apictx = (libsw_memdb_api_t *) api->ctx;

  db = malloc( sizeof( libsw_db_t ));
  if ( ! db )
    goto error;

  db->load_data = libsw_db_load_data;
  db->get_record = libsw_db_get_record;
  db->get_index = libsw_db_get_index;
  db->get_count = libsw_db_get_count;
  db->lock_db = libsw_db_lock_db;
  db->unlock_db = libsw_db_unlock_db;
  db->lock_record = libsw_db_lock_record;
  db->unlock_record = libsw_db_unlock_record;
  db->close = libsw_db_close;


  odb = malloc( sizeof( libsw_obj_db_t ));
  if ( ! odb )
    goto error;

  db->ctx = odb;


  snprintf( odb->filename, sizeof( odb->filename ),
            DBFNPAT, apictx->dir, name );
  odb->count = rcount;
  odb->isize = rsize;

  if (libsw_open_obj_db( apictx, odb,
                         flags,
                         loader, ldctx ) != LIBSW_DB_OK )
  {
    goto error;
  }


  return db;


error:

  libsw_db_close( db );

  return NULL;
}


/**
 * Implements DBA API #libsw_db_api#drop function.
 */
int libsw_db_api_drop( struct libsw_db_api *api, char *name )
{
  char filename[ 256 ];
  libsw_memdb_api_t *apictx = (libsw_memdb_api_t *) api->ctx;

  snprintf( filename, sizeof( filename ), DBFNPAT, apictx->dir, name );

  // TODO: check if the file is locked / open
  if ( unlink ( filename ) != 0 ) {
    return LIBSW_DB_ERROR;
  }

  return LIBSW_DB_OK;
}

/**
 * Implements DBA API #libsw_db_api#close function.
 */
int libsw_db_api_close( struct libsw_db_api *api )
{
  int ret = LIBSW_DB_OK;

  if ( api ) {
    if ( api->ctx ) {
      free( api->ctx );
      api->ctx = NULL;
    }
    free( api );
  }

  return ret;
}

/**
 * Initializes the DB structure to be used with
 * the native memory mapped DB implementation.
 *
 * Configuration parameters are given in the following form: 
 * \c "dir=DBDIR;user=USER;group=GROUP;mode=MODE;". The arguments are
 * as follows:
 *
 * -# \c dir -- the directory to store the memory-mapped DB files;
 * -# \c user -- the file owner user name;
 * -# \c group -- the file group name;
 * -# \c mode -- the file access mode (permissions). Default 0600
 *
 * The parser isn't quite good so all configuration parameters are
 * *mandatory* so far!
 *
 * @param config configuration string
 *
 * @return initialized DBA handle or NULL on error
 */
libsw_db_api_t *libsw_get_memdb_api( const char *config )
{
  libsw_db_api_t *api = NULL;
  libsw_memdb_api_t *ctx = NULL;
  
  ctx = malloc( sizeof( libsw_memdb_api_t ));
  if ( ! ctx )
    goto error;

  if ( config ) {
    char dir[ 128 ], user[ 64 ], group [ 64 ];
    int mode = 0;

    if ( sscanf( config, "dir=%128s;user=%64s;group=%64s;mode=%i;",
                 dir, user, group, &mode ) == 4 )
    {
      struct passwd *pwd;
      struct group *grp;

      snprintf( ctx->dir, sizeof( ctx->dir ), "%s", dir);

      pwd = getpwnam( user );
      if ( ! pwd )
        goto error;
      ctx->uid = pwd->pw_uid;

      grp = getgrnam( group );
      if ( ! grp )
        goto error;
      ctx->gid = grp->gr_gid;

      ctx->mode = mode;
    } else {
      goto error;
    }
  } else {
    snprintf( ctx->dir, sizeof( ctx->dir ), "%s", "/tmp" );
    ctx->uid = getuid();
    ctx->gid = getgid();
    ctx->mode = 0600;
  }

  DBG( "{ dir : \"%s\", uid : %u, gid : %u, mode : %o }",
       ctx->dir, ctx->uid, ctx->gid, ctx->mode );

  api = malloc( sizeof( libsw_db_api_t ));
  if ( ! api )
    goto error;

  api->ctx = ctx;
  api->open = libsw_db_api_open;
  api->drop = libsw_db_api_drop;
  api->close = libsw_db_api_close;


  return api;


 error:

  if ( ctx )
    free( ctx );
  if ( api )
    free( api );

  return NULL;

}
